//	AksPro - ObjectARX module for AutoCAD 2005
//	Copyright (C) 2005 Wiebe Leonid, All Rights Reserved.
//
//	This software is provided 'as-is', without any express or implied
//	warranty.  In no event will the authors be held liable for any
//	damages arising from the use of this software.
//
//	Permission is granted to anyone to use this software for any purpose,
//	including commercial applications, and to alter it and redistribute it
//	freely, subject to the following restrictions:
//
//	1.	The origin of this software must not be misrepresented; you must
//		not claim that you wrote the original software. If you use this
//		software in a product, an acknowledgment in the product
//		documentation would be appreciated but is not required.
//	2.	Altered source versions must be plainly marked as such, and must
//		not be misrepresented as being the original software.
//	3.	This notice may not be removed or altered from any source
//		distribution.

//	AksPro -  ObjectARX  AutoCAD 2005
//	  2005,  
//
//	     ; 
//	   /     
//	     GNU   , 
//	      ; 
//	  2   (  )    .
//
//	   ,    , 
//	        ; 
//	    ,  
//	       . 
//	      GNU.
//
//	        GNU 
//	   ;    ,      
//	(Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA).



//-----------------------------------------------------------------------------
//----- acrxEntryPoint.h
//-----------------------------------------------------------------------------

#include "math.h"

#include "StdAfx.h"
#include "Dlg.h"
#include "SetDlg.h"

#include "axlock.h"
#include "acdocman.h"

//-----------------------------------------------------------------------------
#define szRDS _RXST("W_")

Dlg* dlg = NULL;
SetDlg* sdlg = NULL;
AcDbDatabase *pDb; //    

ACHAR* vers = __T("0.3.2");

BOOL processEntity(AcDbDatabase* db2, AcDbObjectIdArray* basket, AcDbEntity* pEnt, int iType);


bool bSolidIter; //  -   3D 
bool bSDI_changed; //  -   SDI
bool bDragMode; //  -        

// 

PREFS prefFileSet = {0, 0, 20, 0, 1, 0., 45., 90., 1., 1., 1.}; //     
PREFS prefCurSet = {0, 0, 20, 0, 1, 0., 45., 90., 1., 1., 1.}; //     

AcDbVoidPtrArray arSolidLines; //        3D 
AcDbObjectIdArray arSplines; // ,         
AcGePoint3dArray arCenters; //       

std::vector <ACHAR*> vNotProcessed;
std::vector <int> vNotProcessedQty;


int iBodies;
int iEllipses;


//     
AcGePoint3d pBase;

ACHAR sIniFileName[1000];
bool bIniOk = false;


//  Drag 
ads_matrix CMT;

ads_name adsSelSet;

/////////////////////////////

#ifdef DEBUG
void log(ACHAR* str)
{
  FILE* log = fopen("C:/Users//Documents/Visual Studio 2012/Projects/Dimetria/log.log", "a,ccs=UTF-8");
  fwprintf(log, L"%s\n", str);
  fclose(log);
}
#else
#define log(x)
#endif

template <class T, void FreeFunc (T*)>
class AutoFree {
public:
  AutoFree() : m_ptr(0) {}
  AutoFree(T* p) : m_ptr(p) {}
  ~AutoFree() { if (m_ptr) FreeFunc(m_ptr); }
  operator T*& () { return m_ptr; }
  T& operator * () { return *m_ptr; }
  T* operator -> () { return m_ptr; }
  T* ptr() { return m_ptr; }
private:
  T* m_ptr;
};

template <class T>
inline void CloseFunc(T* ptr) { ptr->close(); }

template <class T>
class AutoClose : public AutoFree<T, CloseFunc> {
public:
  AutoClose() : AutoFree() {}
  AutoClose(T* p) : AutoFree(p) {}
};

template <class T>
inline void DeleteFunc(T* ptr) { delete ptr; }

template <class T>
class AutoDelete : public AutoFree<T, DeleteFunc> {
public:
  AutoDelete() : AutoFree() {}
  AutoDelete(T* p) : AutoFree(p) {}
};

Acad::ErrorStatus ZoomExtents(AcApDocument* doc)
{
	return acDocManagerPtr()->sendStringToExecute(doc, _T("_ZOOM\n_EXTENTS\n"), true, false, false);
}

//////////////////////////////////
void addLayer(AcDbDatabase* db, const AcDbEntity* pTemplEnt);

void ApplyTmplProps(AcDbDatabase* db, AcDbEntity* pEnt, const AcDbEntity* pTemplEnt)
{
  if (pTemplEnt != NULL)
  {
    Adesk::UInt16 clr = pTemplEnt->colorIndex();
    pEnt->setColorIndex(clr);

    ACHAR* layer = pTemplEnt->layer();
    if (!bDragMode) addLayer(db, pTemplEnt);
    pEnt->setLayer(layer);

    AcDb::LineWeight lw = pTemplEnt->lineWeight();
    pEnt->setLineWeight(lw);

    double lts = pTemplEnt->linetypeScale();
    pEnt->setLinetypeScale(lts);
	}
}

AcDbObjectId appendAcDbEntity(AcDbDatabase* db, const AcDbEntity* pTemplEnt, AcDbEntity* pEnt)
{
  ApplyTmplProps(db, pEnt, pTemplEnt);

  //if (bDragMode) pLine->setVisibility(AcDb::kInvisible);

  AutoClose<AcDbBlockTable> pBlockTable;
  db->getSymbolTable(pBlockTable, AcDb::kForRead);

  AutoClose<AcDbBlockTableRecord> pBlockTableRecord;
  pBlockTable->getAt(ACDB_MODEL_SPACE, pBlockTableRecord, AcDb::kForWrite);

  AcDbObjectId lineId;
  pBlockTableRecord->appendAcDbEntity(lineId, pEnt);

  return lineId;
}

////////////////////////////////////////////////
//     
bool iseq(
		  double val1,
		  double val2,
      // Must be positive
		  double tol = 0.01
		  )
{
  return (abs(val1-val2) <= tol);
}

class Conic {
public:
  Conic(const AcGePoint3d pt[5])
  {
    double a[5][6];
    for (int i=0; i<5; i++) {
      a[i][0] =  pt[i].x * pt[i].x; //  A
      a[i][1] =  pt[i].x * pt[i].y; //  B
      a[i][2] =  pt[i].y * pt[i].y; //  C
      a[i][3] =  pt[i].x;           //  D
      a[i][4] =  pt[i].y;           //  E
      a[i][5] =  1.0;               //  F
    }

    forward(a);
    backward(a);
  }
  double operator [] (int i) { return x[i]; }
private:
  double x[6];
  
  void sort(double a[5][6], int row)
  {
    int max_row = row;
    for (int i=row+1; i<5; i++)
      if (abs(a[i][row]) > abs(a[max_row][row]))
        max_row = i;

    if (max_row != row)
      for (int j=0; j<6; j++)
      {
        double t = a[row][j];
        a[row][j] = a[max_row][j];
        a[max_row][j] = t;
      }
  }

  void forward(double a[5][6]) //    
  {
    for (int i=1; i<5; i++)
    {
      sort(a, i-1);
      for (int j=i; j<5; j++)
      {
        double m = a[j][i-1]/a[i-1][i-1];
        for (int k=i; k<6; k++)
          a[j][k] -= a[i-1][k]*m;
      }
    }
  }
  void backward(double a[5][6]) //    
  {
    x[5] = 1.0;
    for (int i=4; i>=0; i--)
    {
      double m = 0;
      for (int k=5; k>i;k--)
        m += a[i][k]*x[k];
      x[i] = -m/a[i][i];
    }
  }
};

struct sroots {
  double x[2];
  inline double& operator[] (int i) { return x[i]; }
  inline double Max() { return max(x[0], x[1]); }
};

//    ax^2+bx+c=0
sroots solvSqEq(double a, double b, double c)
{
  double d = b*b - 4*a*c;
  if (d<0)
    throw L"solvSqEq - No roots!";
  double r = sqrt(d);
  sroots result = {{(-b+r)/(2.0*a) , (-b-r)/(2.0*a)}};
  return result;
}

template <typename T>
inline void swap(T& v1, T& v2)
{
  T t = v1;
  v1 = v2;
  v2 = t;
}

const double dblPi = 6.28318530717958647692;

double ang(double a)
{
  while (a<0)
    a += dblPi;
  while (a>dblPi)
    a -= dblPi;
  return a;
}

AcDbEllipse* Ellipse5pt(const AcGePoint3d pt[5])
{
  try {
    Conic conic(pt);
    double a=conic[0], b=conic[1], c=conic[2], d=conic[3], e=conic[4], f=conic[5];

    double x = 4.0*a*c - b*b; // 
    if (x>0) // ,    
    {
      sroots r = solvSqEq(1.0, -(a+c), (a*c)-(0.25*b*b));
      //  
      AcGeVector2d av[2] = { AcGeVector2d(b/2, r[0]-a).normalize(),
                             AcGeVector2d(b/2, r[1]-a).normalize() };
      // 
      double cx = (b*e-2.0*c*d)/x,
             cy = (b*d-2.0*a*e)/x;
      // For radii, solve intersection of axis vectors with Conic Equation:
      // ax^2 + bxy + cy^2 + dx + ey + f = 0  }
      // x = cx + vx(t)                       }- solve for t
      // y = cy + vy(t)                       }
      double rl[2];
      for (int i=0; i<2; i++)
      {
        double vv[2] = {av[i].x*av[i].x, av[i].y*av[i].y};
        double vx = av[i].x,
               vy = av[i].y;
        sroots x = solvSqEq(a*vv[0]+b*vx*vy+c*vv[1],
                            2.0*a*cx*vx+b*(cx*vy+cy*vx)+2.0*c*cy*vy+d*vx+e*vy,
                            a*cx*cx+b*cx*cy+c*cy*cy+d*cx+e*cy+f);
        rl[i] = x.Max();
      }
      if (rl[0]>rl[1])
      {
        swap(rl[0], rl[1]);
        swap(av[0], av[1]);
      }

      resbuf rbFrom, rbTo;
      rbFrom.restype = RTSHORT;
      rbFrom.resval.rint = 1; // From UCS
      rbTo.restype = RTSHORT;
      rbTo.resval.rint = 0;   // To WCS

      ads_point cen = {cx,cy}, cenWCS,
                maj = {av[1].x*rl[1],av[1].y*rl[1]}, majWCS;
      acedTrans(cen, &rbFrom, &rbTo, 0, cenWCS);
      acedTrans(maj, &rbFrom, &rbTo, 0, majWCS);

      return new AcDbEllipse(AcGePoint3d(cenWCS[0],cenWCS[1],cenWCS[2]),
                             AcGeVector3d(0,0,1),
                             AcGeVector3d(majWCS[0],majWCS[1],majWCS[2]),
                             rl[0]/rl[1]);
    }
  }
  catch (wchar_t* err)
  {
    log(L"Exception in Ellipse5pt:");
    log(err);
  }
  catch (...)
  {
    log(L"Unknown exception in Ellipse5pt!");
  }
  return 0; // fail
}

void getPoint(AcGePoint3d* fromPt, AcGePoint3d* resPt, int iType);

AcDbObjectId processEllipse(AcDbDatabase* db,
                            const AcDbEntity* pTemplEnt,
                            const AcGePoint3d pts[5], // axonometry coordinates
                            AcGePoint3d pStart,       // axonometry coordinates
                            AcGePoint3d pEnd,         // axonometry coordinates
                            bool closed)
{
  AutoClose<AcDbEllipse> ellipse = Ellipse5pt(pts);
  if (!ellipse)
  {
    log(L"Error in processEllipse: Ellipse5pt failed!");
    return AcDbObjectId(); // empty
  }

  if (!closed) //  -    
  {
    AcGeVector3d maj = ellipse->majorAxis();
    ads_point mj = {maj.x, maj.y};
    ads_point vc = {0.0, 0.0};
    double angMaj = acutAngle(vc, mj);

    AcGePoint3d cen = ellipse->center();
    ads_point cn = {cen.x, cen.y};
    ads_point vs = {pStart.x, pStart.y};
    double angStart = ang(acutAngle(cn, vs)-angMaj);

    ads_point ve = {pEnd.x, pEnd.y};
    double angEnd = ang(acutAngle(cn, ve)-angMaj);
    //    (  )
    ads_point v1 = {pts[1].x, pts[1].y};
    double angEndAdjusted = ang(angEnd-angStart);
    double ang1 = ang(acutAngle(cn, v1)-angMaj-angStart);
    if (ang1>angEndAdjusted)
      swap(angStart, angEnd);

    ellipse->setStartAngle(angStart);
    ellipse->setEndAngle(angEnd);
  }

  return appendAcDbEntity(db, pTemplEnt, ellipse);
}

AcDbObjectId processArc(AcDbDatabase* db, const AcDbEntity* pTemplEnt, const AcGeCurve3d& pArc, int iType)
{
  AcGePoint3dArray pta;
  pArc.getSamplePoints((pArc.isClosed()) ? 6 : 5, pta);

  AcGePoint3d pStartX, pEndX;
	getPoint(&pta.first(), &pStartX, iType);
	getPoint(&pta.last(), &pEndX, iType);

	//   
	AcGePoint3d pts[5];

	//  
  pts[0] = pStartX;
	for (int i = 1; i < 5; i++)
		getPoint(&pta[i], &pts[i], iType);

  return processEllipse(db,
                        pTemplEnt,
                        pts,     // axonometry coordinates
                        pStartX, // axonometry coordinates
                        pEndX,   // axonometry coordinates
                        pArc.isClosed());
}

AcDbObjectId processEllipticCurve(AcDbDatabase* db, AcDbCurve* pCrv, int iType)
{
  //      dbcurve.h   
  //   getAcGeCurve:
    /// <param name="pGeCurve">Output pointer to an AcGeCurve3d object that is
    /// geometrically identical to this AcDbCurve.  The caller of this function
    /// is responsible for deleting this AcGeCurve3d object.</param>
  //   "delete"       AutoCAD.
  AcGeCurve3d* pArc;
  return pCrv->getAcGeCurve(pArc) == Acad::eOk ? processArc(db, pCrv, *pArc, iType) : AcDbObjectId();
}

inline bool AddToBasket(AcDbObjectIdArray* basket, //   
			                  AcDbObjectId id)
{
  if (!id.isNull()) basket->append(id);
  return !id.isNull();
}

/////////////////////////////
//        
void syncCurSetToFileSet()
{
	prefCurSet.iAksType = prefFileSet.iAksType;
	prefCurSet.iMirror = prefFileSet.iMirror;
	prefCurSet.iSegQty = prefFileSet.iSegQty;
	prefCurSet.iSelElems = prefFileSet.iSelElems;
	prefCurSet.iCreateNewDoc = prefFileSet.iCreateNewDoc; 
	prefCurSet.dAngleX = prefFileSet.dAngleX;
	prefCurSet.dAngleY = prefFileSet.dAngleY;
	prefCurSet.dAngleZ = prefFileSet.dAngleZ;
	prefCurSet.dFactorX = prefFileSet.dFactorX;
	prefCurSet.dFactorY = prefFileSet.dFactorY;
	prefCurSet.dFactorZ = prefFileSet.dFactorZ;
	if (checkSDI(false)) prefCurSet.iCreateNewDoc = 0;
}




/////////////////////////////////////////////
//     
void iterateDb(AcDbDatabase* db)
{

	AutoClose<AcDbBlockTable> pBlkTbl;
	db->getSymbolTable(pBlkTbl, AcDb::kForRead);

	AutoClose<AcDbBlockTableRecord> pBlkTblRcd;
	pBlkTbl->getAt(ACDB_MODEL_SPACE, pBlkTblRcd, AcDb::kForRead);

	AutoDelete<AcDbBlockTableRecordIterator> pBlkTblRcdItr;
	pBlkTblRcd->newIterator(pBlkTblRcdItr);

	acutPrintf(_T("\n>>>>>> iterating model >>>>>>>\n"));

	for (pBlkTblRcdItr->start(); !pBlkTblRcdItr->done(); pBlkTblRcdItr->step())
	{
	  AutoClose<AcDbEntity> pEnt;
		pBlkTblRcdItr->getEntity(pEnt, AcDb::kForRead);
		acutPrintf(_T("\nclassname: %s\n"),pEnt->isA()->name());
	}

	acutPrintf(_T("\n<<<<<< iterating model <<<<<<<\n"));

	acedPostCommandPrompt();
}

//////////////////////////////////////////////////////////////////
//    WCS   
void getPoint(
			  AcGePoint3d* fromPt, 
			  AcGePoint3d* resPt, 
			  double dXAngle, 
			  double dYAngle, 
			  double dZAngle, 
			  double dXDist, 
			  double dYDist, 
			  double dZDist
			  )
{

	AcGePoint3d xp(0.0, 0.0, 0.0);
	AcGePoint3d yp(0.0, 0.0, 0.0);
	AcGePoint3d zp(0.0, 0.0, 0.0);

	int iMirror;
	prefCurSet.iMirror == 0 ? iMirror = 1 : iMirror = -1;

	xp.x = fromPt->x * dXDist * cos(dXAngle * 2 * PI / 360) * iMirror;
	xp.y = fromPt->x * dXDist * sin(dXAngle * 2 * PI / 360);

	yp.x = fromPt->y * dYDist * cos(dYAngle * 2 * PI / 360) * iMirror;
	yp.y = fromPt->y * dYDist * sin(dYAngle * 2 * PI / 360);

	zp.x = fromPt->z * dZDist * cos(dZAngle * 2 * PI / 360) * iMirror;
	zp.y = fromPt->z * dZDist * sin(dZAngle * 2 * PI / 360);

	resPt->x = xp.x + yp.x + zp.x;
	resPt->y = xp.y + yp.y + zp.y;
	resPt->z = 0.0;

}

/////////////////////////////////////////////////////////////////////
void getPoint(
			  AcGePoint3d* fromPt, 
			  AcGePoint3d* resPt, 
			  int iType
			  )
{
  double dXAngle, dYAngle, dZAngle, dXDist, dYDist, dZDist;
  switch (iType) {
  case 0: //  
		dXAngle = 180.+30.;
		dYAngle = 360.-30.;
		dZAngle = 90.;
		dXDist = 1.;
		dYDist = 1.;
		dZDist = 1.;
    break;
  case 1: //  
		dXAngle = 180.+7.17;
		dYAngle = 360.-41.42;
		dZAngle = 90.;
		dXDist = 0.94;
		dYDist = 0.47;
		dZDist = 0.94;
    break;
  case 2: //   
		dXAngle = 180.;
		dYAngle = 360.-45.;
		dZAngle = 90.;
		dXDist = 1.;
		dYDist = 1.;
		dZDist = 1.;
    break;
  case 3: //   
		dXAngle = 180.+60.;
		dYAngle = 360.-30.;
		dZAngle = 90.;
		dXDist = 1.;
		dYDist = 1.;
		dZDist = 1.;
    break;
  case 4: //   
		dXAngle = 180.;
		dYAngle = 360.-45.;
		dZAngle = 90.;
		dXDist = 1.;
		dYDist = 0.5;
		dZDist = 1.;
    break;
  case 5: //  !
		dXAngle = prefCurSet.dAngleX;
		dYAngle = prefCurSet.dAngleY;
		dZAngle = prefCurSet.dAngleZ;
		dXDist = prefCurSet.dFactorX;
		dYDist = prefCurSet.dFactorY;
		dZDist = prefCurSet.dFactorZ;
    break;
  default:
    return;
  }

  getPoint(fromPt, resPt, 
    dXAngle,
    dYAngle,
    dZAngle,
    dXDist,
    dYDist,
    dZDist
    );
}



//////////////////////////////////////
void addLayer(
			  AcDbDatabase* db, 
			  ACHAR* name,
			  int color,
			  bool bIsFrozen,
			  bool bIsLocked,
			  bool bIsOff,
			  bool bIsPlottable,
			  AcDb::LineWeight iLineWeight,
			  bool VPDFLT
			  )
{
  AutoClose<AcDbLayerTable> pLayerTbl;
  if (!(Acad::eOk == db->getSymbolTable(pLayerTbl, AcDb::kForWrite)))
    return;

  if (!pLayerTbl->has(name))
  {
    AutoClose<AcDbLayerTableRecord> pLayerTblRcd = new AcDbLayerTableRecord;
    pLayerTblRcd->setName(name);
    pLayerTblRcd->setIsFrozen(bIsFrozen);// layer to THAWED
    pLayerTblRcd->setIsOff(bIsOff);   // layer to ON
    pLayerTblRcd->setIsPlottable(bIsPlottable);   
    pLayerTblRcd->setVPDFLT(VPDFLT);  // viewport default
    pLayerTblRcd->setLineWeight(iLineWeight);

    AcCmColor clr;
    clr.setColorIndex(color); // set color to red
    pLayerTblRcd->setColor(clr);

    pLayerTblRcd->setIsLocked(0);//  ,      -   

    // For linetype, we need to provide the object ID of
    // the linetype record for the linetype we want to
    // use.  First, we need to get the object ID.
    //AcDbLinetypeTable *pLinetypeTbl;
    //AcDbObjectId ltId;
    //db->getSymbolTable(pLinetypeTbl, AcDb::kForRead);
    //pLinetypeTbl->getAt("CONTINUOUS", ltId);
    //pLinetypeTbl->close();
    //pLayerTblRcd->setLinetypeObjectId(ltId);

    pLayerTbl->add(pLayerTblRcd);
  } 
}

////////////////////////////////////////
bool isOnOffLayer(AcDbEntity* ent)
{
	bool bRetValue = false;

	if (pDb == NULL) return false;
	if (ent == NULL) return false;

	AutoClose<AcDbLayerTable> pLayerTbl;
	if (!(Acad::eOk == pDb->getSymbolTable(pLayerTbl, AcDb::kForRead)))
		return false;

	AcDbObjectId layerId;
	ACHAR* layer = ent->layer();

	if (Acad::eOk == pLayerTbl->getAt(layer, layerId)) //  ID    
	{
		AcDbObject* pLayer;

		//  
		if(Acad::eOk == acdbOpenObject(pLayer, layerId, AcDb::kForRead))
		{
			AutoClose<AcDbLayerTableRecord> layerRc = static_cast<AcDbLayerTableRecord*>(pLayer);
			bRetValue = layerRc->isOff() || layerRc->isFrozen();
			//layerRc->isFrozen()
			//layerRc->isLocked()
			//layerRc->isOff(),
		}
	}

	return bRetValue;
}

//////////////////////////////////////
void addLayer(
			  AcDbDatabase* db, 
			  const AcDbEntity* pTemplEnt
			  )
{

	if (pTemplEnt != NULL)
	{
    bool bEmpty = true;

		//AcDbDatabase* dbt = pTemplEnt->database(); //    
		if (pDb == NULL) return; //   

		AutoClose<AcDbLayerTable> pLayerTbl;
		if (!(Acad::eOk == pDb->getSymbolTable(pLayerTbl, AcDb::kForRead)))
			return;

		AcDbObjectId layerId;
		ACHAR* layer = pTemplEnt->layer();

		if (Acad::eOk == pLayerTbl->getAt(layer, layerId)) //  ID    
		{
			AcDbObject* pLayer;
			//  
			if(Acad::eOk == acdbOpenObject(pLayer, layerId, AcDb::kForRead))
			{
        AutoClose<AcDbLayerTableRecord> layerRc = static_cast<AcDbLayerTableRecord*>(pLayer);
				addLayer(
					db, 
					layer,
					layerRc->color().colorIndex(),
					layerRc->isFrozen(),
					layerRc->isLocked(),
					layerRc->isOff(),
					layerRc->isPlottable(),
					layerRc->lineWeight(),
					layerRc->VPDFLT()
					);
			  bEmpty = false; // !
			}
		}

		if (bEmpty) //   ,   
		{
			addLayer(
				db, 
				layer,
				0,
				false,
				false,
				false,
				true,
				AcDb::kLnWtByLayer,
				true
				);
		}
	}

}


//////////////////////////////////
//    
AcDbObjectId createLine(AcDbDatabase* db, AcDbEntity* pTemplEnt, AcGePoint3d* startPt, AcGePoint3d* endPt)
{
  AutoClose<AcDbLine> pLine = new AcDbLine(*startPt, *endPt);
  return appendAcDbEntity(db, pTemplEnt, pLine);
}

////////////////////////////////////////////////////////
//        
AcDbObjectId createSpline(AcDbDatabase* db, AcDbEntity* pTemplEnt, AcGePoint3dArray* pts, AcGeVector3d* vect)
{
  AutoClose<AcDbSpline> pSpline = new AcDbSpline(*pts, *vect, *vect, 4, 0.0);
  return appendAcDbEntity(db, pTemplEnt, pSpline);
}

//////////////////////////////////
//     
AcDbObjectId createSpline(AcDbDatabase* db, AcDbEntity* pTemplEnt, AcGePoint3dArray* pts)
{
  AutoClose<AcDbSpline> pSpline = new AcDbSpline(*pts);
  return appendAcDbEntity(db, pTemplEnt, pSpline);
}

///////////////////////////////////////
AcDbObjectId createText(
							 AcDbDatabase* db, 
							 AcDbText* pTemplText,
							 CString sText, 
							 double x, 
							 double y, 
							 double ax, 
							 double ay, 
							 double height, 
							 AcDb::TextHorzMode halign, 
							 AcDb::TextVertMode valign, 
							 double dWidthFactor,
							 double rot
							 )
{
  AcGePoint3d startPt(x, y, 0.0);
  AcGePoint3d alPt(ax, ay, 0.0);
  AutoClose<AcDbText> pText = new AcDbText(startPt, sText, AcDbObjectId::kNull, height, 0.0);

  pText->setHorizontalMode(halign);
  pText->setVerticalMode(valign);

  pText->setPosition(startPt);
  pText->setAlignmentPoint(alPt);
  pText->setWidthFactor(dWidthFactor);
  pText->setRotation(rot);

  //AcDb::kTextLeft = 0
  //AcDb::kTextCenter = 1
  //AcDb::kTextRight = 2
  //AcDb::kTextAlign = 3
  //AcDb::kTextMid = 4
  //AcDb::kTextFit = 5

  return appendAcDbEntity(db, pTemplText, pText);
}

///////////////////////////////////////
AcDbObjectId createText(
							 AcDbDatabase* db, 
							 CString sText, 
							 AcDbText* pTemplText,
							 int iType
							 )
{

	AcGePoint3d p = pTemplText->position();
	AcGePoint3d ap = pTemplText->alignmentPoint();
	AcGePoint3d rp;
	AcGePoint3d rap;
	getPoint(&p, &rp, iType);
	getPoint(&ap, &rap, iType);
	double ht = pTemplText->height();
	AcDb::TextHorzMode hm = pTemplText->horizontalMode();
	AcDb::TextVertMode vm = pTemplText->verticalMode();
	double wf = pTemplText->widthFactor();
	double rot = pTemplText->rotation();

	return createText(db, pTemplText, sText, rp.x, rp.y, rap.x, rap.y, ht, hm, vm, wf, rot);
}

///////////////////////////////////////////////////
// for AcDbCurve - AcDbCircle, AcDbEllipse, AcDbArc
AcDbObjectId drawCurve(AcDbDatabase* db, AcDbCurve* pCrv, AcGePoint3d* pTangent, int iType)
{
	AcGePoint3d pStart;
	AcGePoint3d pStartX;
	AcGePoint3d pEnd;
	AcGePoint3d pEndX;
	AcGePoint3d pTangentX;
	AcGeVector3d pVect;
	Acad::ErrorStatus es;

	//  .  . 
	es = pCrv->getStartPoint(pStart);
	es = pCrv->getEndPoint(pEnd);
	getPoint(&pStart, &pStartX, iType);
	getPoint(&pEnd, &pEndX, iType);

	//    ,   
	if (pTangent != NULL)
	{
		getPoint(pTangent, &pTangentX, iType);
		pVect.set(pTangentX.x - pStartX.x, pTangentX.y - pStartX.y, pTangentX.z - pStartX.z);
	}

	double dist = 0.0;

	//   
	if (pCrv->isClosed()) //   
	{
		double param;
		pCrv->getEndParam(param);
		pCrv->getDistAtParam(param, dist);
	}
	else // 
	{
		pCrv->getDistAtPoint(pEnd, dist);
	}

	//  
	dist /= prefCurSet.iSegQty;
	double gap = dist;

	//   
	AcGePoint3dArray pts;

	//  
	pts.append(pStartX);
	for (int i = 0; i < prefCurSet.iSegQty-1; i++)
	{
		AcGePoint3d pt;
		AcGePoint3d pOnCircle;
		es = pCrv->getPointAtDist(gap, pt);
		if (es != Acad::eOk) break;
		getPoint(&pt, &pOnCircle, iType);
		gap += dist;
		pts.append(pOnCircle);
	}
	pts.append(pEndX);

	//  
	return createSpline(db, (AcDbEntity*)pCrv, &pts, &pVect);
}

/////////////////////////////////////
// for AcDbCurve - AcDbCircle, AcDbEllipse, AcDbArc
inline AcDbObjectId drawCurve(AcDbDatabase* db, AcDbCurve* pCrv, int iType)
{
	return drawCurve(db, pCrv, NULL, iType);
}

/////////////////////////////////////
// for polyline arc - AcGeCurve3d
AcDbObjectId drawCurve(
			   AcDbDatabase* db, 
			   AcDbEntity* pTemplEnt, 
			   AcGeCurve3d* pCrv, 
			   AcGePoint3d* pStart, 
			   AcGePoint3d* pEnd, 
			   int iType
			   )
{

	AcGePoint3d pStartX;
	AcGePoint3d pEndX;
	Acad::ErrorStatus es;

	getPoint(pStart, &pStartX, iType);
	getPoint(pEnd, &pEndX, iType);

	double dStartParam = pCrv->paramOf(*pStart);
	double dEndParam = pCrv->paramOf(*pEnd);

	double dist = pCrv->length(dStartParam, dEndParam);
	dist /= prefCurSet.iSegQty;

	double gap = dist;

	AcGePoint3dArray pts;

	pts.append(pStartX);

	for (int i = 0; i < prefCurSet.iSegQty-1; i++)
	{
		AcGePoint3d p = pCrv->evalPoint( dStartParam + ((dEndParam - dStartParam) / prefCurSet.iSegQty * (i+1) ));
		AcGePoint3d pX;
		getPoint(&p, &pX, iType);
		pts.append( pX );
	}

	pts.append(pEndX);

	return createSpline(db, pTemplEnt, &pts);

}

//////////////////////////////////////////////////////////
void processFace(
				 AcDbDatabase* db, // base to save
				 AcDbObjectIdArray* basket, // basket for object ids
				 AcDbEntity* ent, // face to process
				 int iType
				 )
{
	//  !

	AcGePoint3d pStart;
	AcGePoint3d pEnd;
	AcGePoint3d pStartX;
	AcGePoint3d pEndX;
	AcDbFace* face = (AcDbFace*)ent;

	for (int i = 0; i < 4; i++)
	{
		int j;
		if (i < 3) j = i+1; else j = 0;
		face->getVertexAt(i, pStart);
		face->getVertexAt(j, pEnd);

		getPoint(&pStart, &pStartX, iType);
		getPoint(&pEnd, &pEndX, iType);

	  AddToBasket(basket, createLine(db, ent, &pStartX, &pEndX));
	}
}

//////////////////////////////////////////////////////////
AcDbObjectId processClosedSpline(AcDbDatabase* db, AcDbEntity* ent, AcGePoint3d* pCenter, AcGeVector3d* pvNormal, int iType)
{
	AcGePoint3d pStart;
	AcGePoint3d pTangent;

	AcDbCurve* el = (AcDbCurve*)ent;

	el->getStartPoint(pStart);

	//    
	pTangent = pCenter->rotateBy( -PI/2, *pvNormal, pStart);

	return drawCurve(db, el, &pTangent, iType);
}

////////////////////////////////////////////////////////////////////////
//         
int sortFunc(
const void *arg1, 
const void *arg2
)
{
	double* ar1 = (double*)arg1;
	double* ar2 = (double*)arg2;
	if ((double)ar1[1] < (double)ar2[1]) return -1;
	if ((double)ar1[1] > (double)ar2[1]) return 1;
	return 0;
}

////////////////////////////////////////////
//      
// .   
//    
int getTangents(
				 AcDbCurve* crv, 
				 AcGePoint3d* pt, 
				 AcGePoint3dArray* arPoints //        
				 )
{
	arPoints->setLogicalLength(0);

	double dStartParam;
	crv->getStartParam(dStartParam);
	double dEndParam;
	crv->getEndParam(dEndParam);

	double ar[1000][2];

	AcGeVector3d vect2;

	double dGap = (dEndParam - dStartParam)/1000.; //     

	//    
	for(int i = 0; i<1000; i++)
	{
		//    
		Acad::ErrorStatus es = crv->getFirstDeriv(dStartParam, vect2);

		//  
		AcGePoint3d p3;
		crv->getPointAtParam(dStartParam, p3);

		// ,   
		AcDbLine ln(*pt, p3);
		AcGeVector3d vect3;
		es = ln.getFirstDeriv(0.0, vect3);  //   

		//   
		double dAngle = vect3.angleTo(vect2);
		ar[i][0] = dStartParam;
		ar[i][1] = dAngle;

		dStartParam += dGap;
	}

	qsort(ar, 1000, sizeof(double)*2, sortFunc);

	AcGePoint3d ptTn1;
	AcGePoint3d ptTn2;

	crv->getPointAtParam(ar[0][0], ptTn1);
	crv->getPointAtParam(ar[999][0], ptTn2);

	arPoints->append(*pt); //   
	arPoints->append(ptTn1); //   
	arPoints->append(*pt); //   
	arPoints->append(ptTn2); //   

	return 2;
}

//////////////////////////////////
//     
bool drawConeTans(
				  AcDbDatabase* db, 
				  AcDbObjectIdArray* basket, 
				  AcDbEntity* ent,
				  AcDbCurve* cone, 
				  AcGePoint3d* pCenter, 
				  AcGePoint3d* pTop,
				  int iType
				  )
{
	AcGePoint3d p1;
	AcGePoint3d p2;
	getPoint(pCenter, &p1, iType);
	getPoint(pTop, &p2, iType);

	AcGePoint3dArray arPts;
	int iCount = getTangents(cone, &p1, &arPts);

	if (iCount == 2)
	{
		AddToBasket(basket, createLine(db, ent, &(arPts[0]), &(arPts[1])));
		AddToBasket(basket, createLine(db, ent, &(arPts[2]), &(arPts[3])));
	}
	else
	{
		//acutPrintf("\ndrawConeTans: getTangents  %i ,    \n", iCount);
		return false;
	}
	return true;

}

/////////////////////////////////////////////////////
//       
bool tryTangents(
				AcGePoint3d* p1, 
				AcGePoint3d* p2, 
				AcDbCurve* crv, 
				AcGeVector3d* dir, // ,      
				AcDbLine* ar //   ,   ...  
				 )
{

	AcGePoint3d p3(*p1);
	AcGePoint3d p4(*p2);
	AcGePoint3dArray arPts1;
	double angles[4];
	AcDbLine ln2[4]; // 
	AcGeVector3d vct[4];
	int iCount = 0;

	iCount = getTangents(crv, &p3, &arPts1);
	if (iCount == 2)
	{
		ln2[0].setStartPoint(arPts1[0]);
		ln2[0].setEndPoint(arPts1[1]);
		ln2[1].setStartPoint(arPts1[2]);
		ln2[1].setEndPoint(arPts1[3]);
	}
	else
	{
		acutPrintf(_T("\ntryTangents: getTangents  %i \n"), iCount);
		return false;
	}

	iCount = getTangents(crv, &p4, &arPts1);
	if (iCount == 2)
	{
		ln2[2].setStartPoint(arPts1[0]);
		ln2[2].setEndPoint(arPts1[1]);
		ln2[3].setStartPoint(arPts1[2]);
		ln2[3].setEndPoint(arPts1[3]);
	}
	else
	{
		acutPrintf(_T("\ntryTangents: getTangents  %i \n"), iCount);
		return false;
	}

	for (int i = 0; i<4; i++)
	{
		Acad::ErrorStatus es = ln2[i].getFirstDeriv(0.0, vct[i]);
	}


	//AcGeVector3d sum = vct[0] + vct[1] + vct[2] + vct[3];

	for (int i = 0; i<4; i++)
	{
		angles[i] = vct[i].angleTo(*dir);
	}


	for (int i = 0; i<4; i++)
	{
		angles[i] -= PI/2;
		if (angles[i] < 0.) angles[i] *= -1;
	}


	int iIndex1 = 0;
	int iIndex2 = 0;
	double dmax = 0.;


	for (int i = 0; i<4; i++)
	{
		if (angles[i] > dmax)
		{
			dmax = angles[i];
			iIndex1 = i;
		}
	}

	dmax = 0.;
	for (int i = 0; i<4; i++)
	{
		if (iIndex1 == i) continue;
		if (angles[i] > dmax)
		{
			dmax = angles[i];
			iIndex2 = i;
		}
	}


	ar[0].setStartPoint(
		ln2[iIndex1].startPoint() 
		);
	ar[0].setEndPoint(
		ln2[iIndex1].endPoint()
		);
	ar[1].setStartPoint(
		ln2[iIndex2].startPoint()
		);
	ar[1].setEndPoint(
		ln2[iIndex2].endPoint()
		);

	return true;

}


////////////////////////////////////////////////////////////////////
//     ,    
bool drawCylTans(
				 AcDbDatabase* db, 
				 AcDbObjectIdArray* basket, 
				 AcDbEntity* ent, //   
				 AcDbCurve* crv1, 
				 AcDbCurve* crv2, 
				 AcGePoint3d* pCenter1, 
				 AcGePoint3d* pCenter2,
				 int iType
				 )
{
	AcGePoint3d p1;
	AcGePoint3d p2;
	getPoint(pCenter1, &p1, iType);
	getPoint(pCenter2, &p2, iType);
	AcDbLine ln[2]; // 

	AcDbLine dirln(p1, p2);
	AcGeVector3d dir;
	dirln.getFirstDeriv(0.0, dir);


	AcGePoint3d pDist(p1.x - p2.x, p1.y - p2.y, 0.0);
	AcGePoint3d pOut(p1);
	pOut.x += pDist.x * 1000;
	pOut.y += pDist.y * 1000;

	AcGePoint3dArray arPts1;
	AcGePoint3dArray arPts2;

	int iCount = getTangents(crv1, &pOut, &arPts1);

	//createLine(db, ent, &pOut, &p1);
	//createLine(db, ent, &pOut, &p2);

	if (iCount == 2)
	{
		ln[0].setStartPoint(arPts1[0]);
		ln[0].setEndPoint(arPts1[1]);
		ln[1].setStartPoint(arPts1[2]);
		ln[1].setEndPoint(arPts1[3]);
	}
	else
	{
		acutPrintf(_T("\ndrawCylTans: getTangents  %i ,    \n"), iCount);
		return false;
	}

	//id = createLine(db, ent, &(ln[0].startPoint()), &(ln[0].endPoint()));
	//id = createLine(db, ent, &(ln[1].startPoint()), &(ln[1].endPoint()));

	if (!tryTangents(&(arPts1[1]), &(arPts1[3]), crv2, &dir, ln))
	{
		acutPrintf(_T("\ndrawCylTans: tryTangents   1 \n"), iCount);
		return false;
	}
	if (!tryTangents(&(ln[0].endPoint()), &(ln[1].endPoint()), crv1, &dir, ln))
	{
		acutPrintf(_T("\ndrawCylTans: tryTangents   2 \n"), iCount);
		return false;
	}
	if (!tryTangents(&(ln[0].endPoint()), &(ln[1].endPoint()), crv2, &dir, ln))
	{
		acutPrintf(_T("\ndrawCylTans: tryTangents   3 \n"), iCount);
		return false;
	}

	AddToBasket(basket, createLine(db, ent, &(ln[0].startPoint()), &(ln[0].endPoint())));
	AddToBasket(basket, createLine(db, ent, &(ln[1].startPoint()), &(ln[1].endPoint())));

	return true;

}


///////////////////////////////////////////////////////////////////
void processTangents(
					 AcDbDatabase* db, 
					 AcDbObjectIdArray* basket, //   
					 AcDbEntity* ent, //   
					 AcDbObjectIdArray arSplines, //     
					 AcGePoint3dArray arCenters, //   (    )
					 int iType
					 )
{
  // 
  if (arSplines.length() == 1)
  {
    AutoClose<AcDbEntity> pEnt;
    if (acdbOpenAcDbEntity(pEnt, arSplines[0], AcDb::kForRead) == Acad::eOk)
    {
      AcDbCurve* crv = static_cast<AcDbCurve*>(pEnt.ptr());
      drawConeTans(db, basket, ent, crv, &(arCenters[0]), &(arCenters[1]), iType);
    }
  }
  //    
  else if (arSplines.length() == 2)
  {
    Acad::ErrorStatus es1;
    Acad::ErrorStatus es2;
    AutoClose<AcDbEntity> pEnt1;
    AutoClose<AcDbEntity> pEnt2;
    es1 = acdbOpenAcDbEntity(pEnt1, arSplines[0], AcDb::kForRead);
    es2 = acdbOpenAcDbEntity(pEnt2, arSplines[1], AcDb::kForRead);

    if (es1 == Acad::eOk && es2 == Acad::eOk)
    {
      AcDbCurve* crv1 = static_cast<AcDbCurve*>(pEnt1.ptr());
      AcDbCurve* crv2 = static_cast<AcDbCurve*>(pEnt2.ptr());
      drawCylTans(db, basket, ent, crv1, crv2, &(arCenters[0]), &(arCenters[1]), iType);
    }
  }
}

//////////////////////////////////////////////////////////////
//   
// .   
int explodeEnt(
			   AcDbDatabase* db, 
			   AcDbObjectIdArray* basket, //   
			   AcDbEntity* ent, 
			   int iType
			   )
{

	AcDbVoidPtrArray entitySet;
	Acad::ErrorStatus es = ent->explode(entitySet);

	bool bFound = false;
	int iNotProcessed = 0;


	//     3D  
	if (AcDb3dSolid::cast(ent))
	{
		iBodies = 0;
		iEllipses = 0;
		bSolidIter = true;
	}

	//   
	for (int i = 0; i < entitySet.length(); i++) 
	{
		AcDbEntity* pIntEnt = (AcDbEntity*)entitySet[i];
		//acutPrintf("\nclassname: %s\n",(pIntEnt->isA())->name());


		if (
			AcDb3dSolid::cast(pIntEnt) ||
			AcDbRegion::cast(pIntEnt)
			)
		{
			iNotProcessed += explodeEnt(db, basket, pIntEnt, iType);
		}
		else if (AcDbFace::cast(pIntEnt))
		{
			processFace(db, basket, pIntEnt, iType);
		}
		else if (AcDbBody::cast(pIntEnt))
		{
			//  ...
			iBodies++;

			//explodeEnt(db, basket, pIntEnt, iType);

			//     ,  
			if (bSolidIter)
			{
				AcDbIntArray osnapModes;
				AcDbIntArray geomIds;
				pIntEnt->getGripPoints(arCenters, osnapModes, geomIds);
			}
			else
			{
				iNotProcessed++;
			}

		}
		else if (AcDbLine::cast(pIntEnt))
		{
			AcDbLine* lin = (AcDbLine*)pIntEnt;

			AcGePoint3d pStart;
			AcGePoint3d pEnd;
			AcGePoint3d pStartX;
			AcGePoint3d pEndX;

			pStart = lin->startPoint();
			pEnd = lin->endPoint();

			//   
			if (bSolidIter) 
			{
				AcGeTol tol;
				tol.setEqualPoint(0.01);
				bFound = false;
				// ,      
				for (
					int i = 0; 
					i < arSolidLines.length(); 
					i++
					)
				{
					AcGePoint3d p1 = 
						((AcDbLine*)arSolidLines[i])->startPoint();
					AcGePoint3d p2 = 
						((AcDbLine*)arSolidLines[i])->endPoint();
					if (
						(((iseq(p1.x, pStart.x)) && 
						(iseq(p1.y, pStart.y)) && 
						(iseq(p1.z, pStart.z))) &&
						((iseq(p2.x, pEnd.x)) && 
						(iseq(p2.y, pEnd.y)) && 
						(iseq(p2.z, pEnd.z)))) ||
						(((iseq(p1.x, pEnd.x)) && 
						(iseq(p1.y, pEnd.y)) && 
						(iseq(p1.z, pEnd.z))) &&
						((iseq(p2.x, pStart.x)) && 
						(iseq(p2.y, pStart.y)) && 
						(iseq(p2.z, pStart.z))))
						//(p1.isEqualTo(pStart, tol) && 
						//p2.isEqualTo(pEnd, tol)) ||
						//(p1.isEqualTo(pEnd, tol) && 
						//p2.isEqualTo(pStart, tol))
						)
					{
						bFound = true;
						break;
					}
				}
				//  ,     
				if (bFound) continue;
			}

			getPoint(&pStart, &pStartX, iType);
			getPoint(&pEnd, &pEndX, iType);

			AddToBasket(basket, createLine(db, pIntEnt, &pStartX, &pEndX));

			if (bSolidIter) 
			{
				//  ,        
				AcDbLine* lnnew = new AcDbLine(lin->startPoint(), lin->endPoint());
				arSolidLines.append(lnnew);
			}
		}
		else if (AcDbCircle* el = AcDbCircle::cast(pIntEnt))
		{
			iEllipses++;
      AcDbObjectId id = processEllipticCurve(db, el, iType);
			if (id.isNull())
  			id = processClosedSpline(db, pIntEnt, &(el->center()), &(el->normal()), iType);
			if (!id.isNull())
			{
				if (bSolidIter) arSplines.append(id);
				basket->append(id);
			}
		}
		else if (AcDbEllipse* el = AcDbEllipse::cast(pIntEnt))
		{
			iEllipses++;
      AcDbObjectId id = processEllipticCurve(db, el, iType);
			if (id.isNull())
			  id = processClosedSpline(db, pIntEnt, &(el->center()), &(el->normal()), iType);
			if (!id.isNull())
			{
				if (bSolidIter) arSplines.append(id);
				basket->append(id);
			}
		}
		else if (AcDbCurve* curv = AcDbArc::cast(pIntEnt))
		{
      if (!AddToBasket(basket, processEllipticCurve(db, curv, iType)))
			  AddToBasket(basket, drawCurve(db, curv, iType));
		}
		else
		{
			processEntity(db, basket, pIntEnt, iType);
			//acutPrintf("\n      %s\n",(pIntEnt->isA())->name());
			//iNotProcessed++;
		}
	}


	for (int i = 0; i < entitySet.length(); i++) 
		delete (AcRxObject*)entitySet[i];

	//    3D 
	if (AcDb3dSolid::cast(ent))
	{
		if ((iBodies == 1 && iEllipses == 2) || (iBodies == 1 && iEllipses == 1))
		{
			processTangents(db, basket, (AcDbEntity*)ent, arSplines, arCenters, iType);
		}

		//   
		bSolidIter = false;
		for (int i = 0; i < arSolidLines.length(); i++)
			delete (AcDbLine*)arSolidLines[i];
		
		arSolidLines.setLogicalLength(0);

		arSplines.setLogicalLength(0);
		arCenters.setLogicalLength(0);
	}

	return iNotProcessed;

}

///////////////////////////////////////////////
//      
BOOL processEntity(AcDbDatabase* db2, AcDbObjectIdArray* basket, AcDbEntity* pEnt, int iType)
{
  AcGePoint3d pStart;
  AcGePoint3d pEnd;
  AcGePoint3d pStartX;
  AcGePoint3d pEndX;

  if (AcDbLine* line = AcDbLine::cast(pEnt))
  {
    pStart = line->startPoint();
    pEnd = line->endPoint();

    getPoint(&pStart, &pStartX, iType);
    getPoint(&pEnd, &pEndX, iType);

    AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));
  }
  else if (AcDbPolyline* lin = AcDbPolyline::cast(pEnt))
  {
    for (int i = 0; i < lin->numVerts(); i++)
    {
      int j;
      if (i == lin->numVerts() - 1) 
        if (lin->isClosed()) j = 0; else break;
      else
        j = i + 1;

      lin->getPointAt(i, pStart);
      lin->getPointAt(j, pEnd);

      AcDbPolyline::SegType st = lin->segType(i);
      //kLine   This is a non-zero length line segment. 
      //kArc   This is a non-zero length arc segment. 
      //kCoincident   This is a zero-length segment. 
      //kPoint   This is a single vertex polyline. 
      //kEmpty   This entity has no vertices. 

      if (st == AcDbPolyline::kLine)
      {
        getPoint(&pStart, &pStartX, iType);
        getPoint(&pEnd, &pEndX, iType);
        AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));
      }
      else if (st == AcDbPolyline::kArc)
      {
        AcGeCircArc3d arc;
        lin->getArcSegAt(i, arc);
        if (!AddToBasket(basket, processArc(db2, lin, arc, iType)))
          AddToBasket(basket, drawCurve(db2, lin, &arc, &pStart, &pEnd, iType));
      }
      else
        continue;
    }
  }
  else if (AcDb3dPolyline* pPline = AcDb3dPolyline::cast(pEnt))
  {
    //		AcDb::Poly3dType pt = pPline->polyType();
    AutoDelete<AcDbObjectIterator> pVertIter= pPline->vertexIterator();

    for (int vertexNumber = 0; !pVertIter->done();
      vertexNumber++, pVertIter->step())
    {
      AcDbObjectId vertexObjId = pVertIter->objectId();
      AutoClose<AcDb3dPolylineVertex> pVertex;
      pPline->openVertex(pVertex, vertexObjId, AcDb::kForRead);
      pEnd = pVertex->position();

      if (vertexNumber)
      {
        getPoint(&pStart, &pStartX, iType);
        getPoint(&pEnd, &pEndX, iType);
        AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));
      }
      pStart.set(pEnd.x, pEnd.y, pEnd.z);
    }
  }
  else if (AcDb2dPolyline* pPline = AcDb2dPolyline::cast(pEnt))
  {
    //		AcDb::Poly2dType pt = pPline->polyType();
    AutoDelete<AcDbObjectIterator> pVertIter= pPline->vertexIterator();

    for (int vertexNumber = 0; !pVertIter->done();
      vertexNumber++, pVertIter->step())
    {
      AcDbObjectId vertexObjId = pVertIter->objectId();
      AutoClose<AcDb2dVertex> pVertex;
      pPline->openVertex(pVertex, vertexObjId, AcDb::kForRead);
      pEnd = pVertex->position();

      if (vertexNumber)
      {
        getPoint(&pStart, &pStartX, iType);
        getPoint(&pEnd, &pEndX, iType);
        AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));
      }
      pStart.set(pEnd.x, pEnd.y, pEnd.z);
    }
  }
  else if (AcDbCircle* el = AcDbCircle::cast(pEnt))
  {
    if (!AddToBasket(basket, processEllipticCurve(db2, el, iType)))
      AddToBasket(basket, processClosedSpline(db2, pEnt, &(el->center()), &(el->normal()), iType));
  }
  else if (AcDbEllipse* el = AcDbEllipse::cast(pEnt))
  {
    if (!AddToBasket(basket, processEllipticCurve(db2, el, iType)))
      AddToBasket(basket, processClosedSpline(db2, pEnt, &(el->center()), &(el->normal()), iType));
  }
  else if (AcDbCurve* cur = AcDbArc::cast(pEnt))
  {
    if (!AddToBasket(basket, processEllipticCurve(db2, cur, iType)))
      AddToBasket(basket, drawCurve(db2, cur, iType));
  }
  else if (
    AcDb3dSolid::cast(pEnt) ||
    AcDbRegion::cast(pEnt) ||
    AcDbHatch::cast(pEnt) ||
    AcDbBlockReference::cast(pEnt) ||
    AcDbPolygonMesh::cast(pEnt) ||
    //AcDbBody::cast(pEnt) ||
    AcDbPolyFaceMesh::cast(pEnt)
    )
  {
    if (explodeEnt(db2, basket, pEnt, iType)) return FALSE;
  }
  else if (AcDbSolid* face = AcDbSolid::cast(pEnt))
  {
    //return FALSE;

    face->getPointAt(0, pStart);
    face->getPointAt(1, pEnd);
    getPoint(&pStart, &pStartX, iType);
    getPoint(&pEnd, &pEndX, iType);
    AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));

    face->getPointAt(1, pStart);
    face->getPointAt(3, pEnd);
    getPoint(&pStart, &pStartX, iType);
    getPoint(&pEnd, &pEndX, iType);
    AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));

    face->getPointAt(3, pStart);
    face->getPointAt(2, pEnd);
    getPoint(&pStart, &pStartX, iType);
    getPoint(&pEnd, &pEndX, iType);
    AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));

    face->getPointAt(2, pStart);
    face->getPointAt(0, pEnd);
    getPoint(&pStart, &pStartX, iType);
    getPoint(&pEnd, &pEndX, iType);
    AddToBasket(basket, createLine(db2, pEnt, &pStartX, &pEndX));
  }
  else if (AcDbFace::cast(pEnt))
  {
    processFace(db2, basket, pEnt, iType);
  }
  else if (AcDbText* text = AcDbText::cast(pEnt))
  {
    AddToBasket(basket, createText(db2, text->textString(), ((AcDbText*)pEnt), iType));
  }
  else
  {
    const ACHAR* ename = pEnt->isA()->name();
    bool bFound = false;
    for (int i = 0; i < vNotProcessed.size(); i++)
    {
      if (_tcscmp(vNotProcessed[i], ename) == 0)
      {
        vNotProcessedQty[i]++;
        bFound = true;
      }
    }

    if (!bFound)
    {
      ACHAR* ar = new ACHAR[100];
      _tcsncpy(ar, ename, 100);
      vNotProcessed.push_back(ar);
      vNotProcessedQty.push_back(1);
    }
    //acutPrintf("\n   %s\n",(pEnt->isA())->name());
  }
  return TRUE;

}

/////////////////////////////////
void printError(int iNumErrors)
{
  const ACHAR* fmtStr;
  switch (iNumErrors % 10) {
  case 1:
    fmtStr = _T("\n      %i \n");
    break;
  case 2: case 3: case 4:
    fmtStr = _T("\n      %i \n");
    break;
  default:
    fmtStr = _T("\n      %i \n");
    break;
  }
  acutPrintf(fmtStr, iNumErrors);
}

///////////////////////////////////////
void reportNotProcessed()
{
	for (int i = 0; i < vNotProcessed.size(); i++)
	{
		acutPrintf(_T("\n  %s - %i \n"), vNotProcessed[i], vNotProcessedQty[i]);
		ACHAR* ar = vNotProcessed[i];
		delete[] ar;
	}
	vNotProcessed.clear();
	vNotProcessedQty.clear();

	acedPostCommandPrompt();
}

/////////////////////////////////////////////////////////////////////
void processEntities(AcDbDatabase* db2, AcDbObjectIdArray* basket, ads_name ss, long length, int iType)
{

  ads_name     ent;
  AcDbObjectId id;

  int iNumErrors = 0;

  acedSetStatusBarProgressMeter(_T("     "), 0, length);

  for (long i = 0; i < length; i++) 
  {

    acedSetStatusBarProgressMeterPos(i);

    if (RTNORM != acedSSName( ss, i, ent ))
      continue;
    if (Acad::eOk != acdbGetObjectId( id, ent ))
      continue;
    AutoClose<AcDbEntity> pEnt;
    if (Acad::eOk != acdbOpenAcDbEntity( pEnt, id, AcDb::kForRead ))
      continue;

    if (!isOnOffLayer(pEnt))
    {
      if (!processEntity(db2, basket, pEnt, iType))
        iNumErrors++;
    }
  }

  acedRestoreStatusBar();
}

/////////////////////////////////////////////////////////////////////
void processEntities(ads_name ss, long length, int iType)
{
	pDb = acdbHostApplicationServices()->workingDatabase();

	acDocManager->appContextNewDocument(_T("acad.dwt"));
	AcApDocument* pNewDoc = acDocManagerPtr()->curDocument();  
	acDocManagerPtr()->lockDocument(pNewDoc);
	AcDbDatabase* db2 = pNewDoc->database();

	AcDbObjectIdArray ids;

	processEntities(db2, &ids, ss, length, iType);

	Acad::ErrorStatus es = ZoomExtents(pNewDoc);
}

////////////////////////////////
void processDatabase(AcDbDatabase* db, AcDbDatabase* db2, int iType)
{

	int iEntities = 0;
	int iNumErrors = 0;
	AcDbObjectIdArray ids; //     

	// let's go iterate elements
	AutoClose<AcDbBlockTable> pBlkTbl;
	db->getSymbolTable(pBlkTbl, AcDb::kForRead);

	AutoClose<AcDbBlockTableRecord> pBlkTblRcd;
	pBlkTbl->getAt(ACDB_MODEL_SPACE, pBlkTblRcd, AcDb::kForRead);

	AutoDelete<AcDbBlockTableRecordIterator> pBlkTblRcdItr;
	pBlkTblRcd->newIterator(pBlkTblRcdItr);

  for (; !pBlkTblRcdItr->done();	pBlkTblRcdItr->step())
			iEntities++;

	if (iEntities == 0)
	{
		acutPrintf(_T("\n    !\n"));
		acedPostCommandPrompt();
		return;
	}

	if (dlg != NULL) dlg->m_Progress.SetRange(1, iEntities);
	acedSetStatusBarProgressMeter(_T("     "), 1, iEntities);

	for (pBlkTblRcdItr->start(); !pBlkTblRcdItr->done();	pBlkTblRcdItr->step())
	{
	  AutoClose<AcDbEntity> pEnt;
		pBlkTblRcdItr->getEntity(pEnt, AcDb::kForRead);
		//acutPrintf("\nclassname: %s\n",(pEnt->isA())->name());

		if (dlg != NULL) dlg->m_Progress.OffsetPos(1);
		acedSetStatusBarProgressMeterPos(-1); // Negative is offset (relative)
		
		if (!isOnOffLayer(pEnt))
		{
			if (!processEntity(db2, &ids, pEnt, iType))
				iNumErrors++;
		}
	}

	if (dlg != NULL) dlg->m_Progress.SetPos(1);
	acedRestoreStatusBar();

}

////////////////////////////////
void processDatabase(int iType)
{
	pDb = acdbHostApplicationServices()->workingDatabase();

	// new document
	Acad::ErrorStatus es = eOk;

	es = acDocManager->appContextNewDocument(_T("acad.dwt"));

	AcApDocument* pNewDoc=acDocManagerPtr()->curDocument();  
	es = acDocManagerPtr()->lockDocument(pNewDoc);
	AcDbDatabase* db2 = pNewDoc->database();

	processDatabase(pDb, db2, iType);

	es = ZoomExtents(pNewDoc);

}


/////////////////////////////////
BOOL checkSDI(bool bMes = true)
{
	struct resbuf rb; 
	acedGetVar(_T("SDI"), &rb);
	if (!bMes) return (BOOL)rb.resval.rint;

	bSDI_changed = false;

	if (rb.resval.rint != 0)
	{
		acutPrintf(_T("\n      SDI    0\n"));
		if (MessageBox(NULL, _T("      SDI    0\n         \n  ?"), _T("AksPro"), MB_OKCANCEL) == IDOK)
		{
			rb.resval.rint = 0;
			if (acedSetVar(_T("SDI"), &rb) != RTNORM)
			{
				acutPrintf(_T(",    SDI \n"));
				return FALSE;
			}
			else
				bSDI_changed = true;
		}
		else
		{
			acutPrintf(_T("\n AKSPRO    SDI = 0\n"));
			return FALSE;
		}
	}

	return TRUE;
}

//////////////////////////////////////
void newSyncDocHelper( void *pData)
{
	//   
    AcApDocument* pDoc = acDocManager->curDocument();

	if (prefCurSet.iSelElems)
	{
		if (RTNORM != acedSSGet( NULL, NULL, NULL, NULL, adsSelSet ))
		{
			acutPrintf( _T("\n \n") );
			acedPostCommandPrompt();
			return;
		}
	}
	

	//   
    if (acDocManager->isApplicationContext()) 
        acDocManager->appContextNewDocument((const ACHAR *)pData);

	//  
	AcApDocument* pNewDoc = acDocManager->curDocument();

	// 
	Acad::ErrorStatus es = acDocManagerPtr()->lockDocument(pNewDoc);

	//  
	AcDbDatabase* db2 = pNewDoc->database();

	long length;
	if ((RTNORM == acedSSLength( adsSelSet, &length )) && (0 != length))
	{
		//ads_name ss;
		//if (RTNORM != acedSSGet( NULL, NULL, NULL, NULL, ss ))
		//	acutPrintf( "\n \n" );
		//acedPostCommandPrompt();
		AcDbObjectIdArray ids;

		processEntities(db2, &ids, adsSelSet, length, prefCurSet.iAksType);

		acedSSFree( adsSelSet );
	}
	else
	{
		processDatabase(pDoc->database(), db2, prefCurSet.iAksType);
	}

	reportNotProcessed();

	es = ZoomExtents(pNewDoc);
}

//////////////////////////////////////
//void newSyncDoc()
//{
//    static char pData[] = "acad.dwt";
//
//    AcApDocument* pDoc = acDocManager->curDocument();
//    if (pDoc) 
//        acDocManager->executeInApplicationContext(newSyncDocHelper, (void *)pData);
//}

//////////////////////////////////////
int drag(ads_point pt, ads_matrix mat)
{
	ads_point       tp;
	int             axis;

	AcGePoint3d pPt = asPnt3d(pt);

	acdbUcs2Wcs(asDblArray(pBase), asDblArray(pBase),false);
	acdbUcs2Wcs(asDblArray(pPt), asDblArray(pPt),false);

	AcGeVector3d v = pPt - pBase;
	AcGeMatrix3d gMat; 
	gMat.setToTranslation(v);

	memcpy(mat, gMat.entry, sizeof(ads_matrix));
	memcpy(CMT, mat, sizeof(ads_matrix));

	return RTNORM;
}

////////////////////////////
int goAks(
		  int iAxType = prefCurSet.iAksType,
		  int iSel = prefCurSet.iSelElems,
		  bool bNewDoc = (bool)prefCurSet.iCreateNewDoc,
		  double dAnX = prefCurSet.dAngleX,
		  double dAnY = prefCurSet.dAngleY,
		  double dAnZ = prefCurSet.dAngleZ,
		  double dFcX = prefCurSet.dFactorX,
		  double dFcY = prefCurSet.dFactorY,
		  double dFcZ = prefCurSet.dFactorZ
		 )
{
	bDragMode = !bNewDoc;

	if (bNewDoc)
	{
		pDb = acdbHostApplicationServices()->workingDatabase();
		static ACHAR pData[] = _T("acad.dwt");
		AcApDocument* pDoc = acDocManager->curDocument();
		if (pDoc)
			acDocManager->executeInApplicationContext(newSyncDocHelper, (void *)pData);
	}
	else
	{
		//   
		AcApDocument* pDoc = acDocManager->curDocument();
		long length;

		if (iSel)
		{
			if (RTNORM != acedSSGet( NULL, NULL, NULL, NULL, adsSelSet ))
			{
				acutPrintf( _T("\n \n") );
				acedPostCommandPrompt();
				return 0;
			}
		}
		//    ,   
		else if (!(RTNORM == acedSSLength( adsSelSet, &length ) && length > 0))
		{
			if (RTNORM != acedSSGet( _T("_A"), NULL, NULL, NULL, adsSelSet ))
				return 0;
		}


		//  
		AcDbDatabase* db = pDoc->database();

		AcDbObjectIdArray basket_ids;

		if (!(RTNORM != acedSSLength( adsSelSet, &length ) || 0 == length))
			processEntities(db, &basket_ids, adsSelSet, length, prefCurSet.iAksType);

		acedSSFree( adsSelSet );

		ads_name ss2;

		acedSSAdd(NULL, NULL, ss2);

		AcDbExtents ext;

		for (int i = 0; i<basket_ids.length(); i++)
		{
			Acad::ErrorStatus es;
			AutoClose<AcDbEntity> pEnt;
			es = acdbOpenAcDbEntity(pEnt, basket_ids[i], AcDb::kForRead);
			if (es == Acad::eOk)
			{
				AcDbExtents curext;
				pEnt->getGeomExtents(curext);
				ext.addExt(curext);
			}

			ads_name name;
			es = acdbGetAdsName(name, basket_ids[i]);
			if (es == Acad::eOk)
				acedSSAdd(name, ss2, ss2);
		}

		pBase.set(
			(ext.minPoint().x + ext.maxPoint().x)/2, 
			(ext.minPoint().y + ext.maxPoint().y)/2, 
			(ext.minPoint().z + ext.maxPoint().z)/2
			);

		if (!(RTNORM != acedSSLength( ss2, &length ) || 0 == length))
		{
			int rc; 
			ads_point return_pt; 

			memcpy(CMT, AcGeMatrix3d::kIdentity.entry, sizeof(ads_matrix));

			if (acedDragGen(ss2,_T("\n    :\n"), 0, drag, return_pt) == RTNORM)
			{
				//    CMT   p
				ads_matrix mat;
				drag(return_pt, mat);
				acedXformSS(ss2,CMT);
			}
			else
			{

				//   ,    
				acedSetStatusBarProgressMeter(_T("   ...  "), 0, basket_ids.length());
				for (int i = 0; i<basket_ids.length(); i++)
				{
					ads_name name;
					Acad::ErrorStatus es = acdbGetAdsName(name, basket_ids[i]);
					acedSetStatusBarProgressMeterPos(i);
					if (es == Acad::eOk)
						acdbEntDel(name);

				}
				acedRestoreStatusBar();
			}
		}

		acedSSFree( ss2 );

	}


	return 0;

}


//////////////////////////////////////////////
//    
int goAks(struct resbuf *pArgs, bool bNewDoc = (bool)prefCurSet.iCreateNewDoc, int iSel = prefCurSet.iSelElems)
{
	prefCurSet.iSelElems = iSel;
	prefCurSet.iCreateNewDoc = (int)bNewDoc;

	bool bError = false;
	bool bAngles = false;

	if (pArgs == NULL) 
		bError = true;
	else
	{
		//     
		if (pArgs->restype == RTSHORT && pArgs->rbnext == NULL) 
		{ 
			prefCurSet.iAksType = pArgs->resval.rint;
			if (prefCurSet.iAksType < 0 || prefCurSet.iAksType > 4) 
				bError = true;
		} 
		else if (pArgs->restype == RTREAL || pArgs->restype == RTSHORT)
		{
			if (pArgs->restype == RTREAL) prefCurSet.dAngleX = pArgs->resval.rreal;
			else if (pArgs->restype == RTSHORT) prefCurSet.dAngleX = pArgs->resval.rint;
			else bError = true;

			if (pArgs->rbnext == NULL)
				bError = true;
			else
			{
				pArgs = pArgs->rbnext;
				if (pArgs->restype == RTREAL) prefCurSet.dAngleY = pArgs->resval.rreal;
				else if (pArgs->restype == RTSHORT) prefCurSet.dAngleY = pArgs->resval.rint;
				else bError = true;

				if (pArgs->rbnext == NULL)
					bError = true;
				else
				{
					pArgs = pArgs->rbnext;
					if (pArgs->restype == RTREAL) prefCurSet.dAngleZ = pArgs->resval.rreal;
					else if (pArgs->restype == RTSHORT) prefCurSet.dAngleZ = pArgs->resval.rint;
					else bError = true;
					bAngles = true;
				}
			}
		} 
		else if (pArgs->restype == RTPICKS)
		{
			prefCurSet.iSelElems = 0;
			ads_name_set(pArgs->resval.rlname, adsSelSet);
			long length;
			if ((RTNORM != acedSSLength( adsSelSet, &length ) || 0 == length))
				return 0;
		}
		else
		{
			return 0;
		}
	}

	
	if (bError)
	{ 
		acutPrintf(_T("\n   :\n"));
		acutPrintf(_T("\n    -    0  4     2.317-69\n"));
		acutPrintf(_T("\n    -       X, Y, Z\n"));
		return RTERROR;
	} 
	else if (bAngles)
	{
		prefCurSet.iAksType = 5; //   
	}


	return goAks();
}


/////////////////////////////////////
bool getSet(char* set, double* retval)
{
	if (set == NULL) return false;
	set = strchr(set, '='); 
	if (set == NULL) return false;
	set++; 
	*retval = atof(set);
	return true;
}
//////////////////////////////////
bool getSet(char* set, int* retval)
{
	if (set == NULL) return false;
	set = strchr(set, '='); 
	if (set == NULL) return false;
	set++; 
	*retval = atoi(set);
	return true;
}


///////////////////////////
void saveSettings()
{
	if (!bIniOk) return; // acedGetAppName()   

	std::fstream file;
	file.open( sIniFileName, std::ios_base::out | std::ios_base::trunc );

	char val[100];

	sprintf(val, "%i", prefFileSet.iAksType);		file << "AksonometryType=" << val << std::endl; 
	sprintf(val, "%.2f", prefFileSet.dAngleX);		file << "AxisX=" << val << std::endl; 
	sprintf(val, "%.2f", prefFileSet.dAngleY);		file << "AxisY=" << val << std::endl; 
	sprintf(val, "%.2f", prefFileSet.dAngleZ);		file << "AxisZ=" << val << std::endl; 
	sprintf(val, "%.2f", prefFileSet.dFactorX);		file << "FactorX=" << val << std::endl; 
	sprintf(val, "%.2f", prefFileSet.dFactorY);		file << "FactorY=" << val << std::endl; 
	sprintf(val, "%.2f", prefFileSet.dFactorZ);		file << "FactorZ=" << val << std::endl; 
	sprintf(val, "%i", prefFileSet.iSegQty);		file << "SegmentsQuantity=" << val << std::endl; 
	sprintf(val, "%i", prefFileSet.iSelElems);		file << "SelectElements=" << val << std::endl; 
	sprintf(val, "%i", prefFileSet.iMirror);		file << "MirrorElements=" << val << std::endl; 
	sprintf(val, "%i", prefFileSet.iCreateNewDoc);	file << "CreateNewDocument=" << val << std::endl; 
	file.close( );

//	__asm nop //       !... :)
}

//////////////////////
bool getIniFileName()
{
	// acedGetAppName()   

	ACHAR* appname = acedGetAppName();

	if (appname == NULL) return false;

	_tcscpy(sIniFileName, appname);
	ACHAR* s = sIniFileName;
	s = _tcsrchr(s, '\\');
	if (s == NULL) return false;

	s++;
	_tcscpy(s, _T("W_AksPro.ini\0"));

	bIniOk = true;

	return true;
}


///////////////////////////
void getSettings()
{

	char AksonometryType[100];

	char AxisX[100];
	char AxisY[100];
	char AxisZ[100];

	char FactorX[100];
	char FactorY[100];
	char FactorZ[100];

	char SegmentsQuantity[100];

	char SelectElements[100];

	char MirrorElements[100];
	char CreateNewDocument[100];

	getIniFileName();
	if (!bIniOk) return; // acedGetAppName()   


	std::ifstream file;

	file.open( sIniFileName );


	if (file.fail())
	{
		acutPrintf(_T("\n    - %s\n"), sIniFileName);
		return;
	}

	bool bErr = false;

	file >> AksonometryType; 
	if (!getSet(AksonometryType, &(prefFileSet.iAksType))) bErr = true;
	file >> AxisX; 
	if (!getSet(AxisX, &(prefFileSet.dAngleX))) bErr = true;
	file >> AxisY; 
	if (!getSet(AxisY, &(prefFileSet.dAngleY))) bErr = true;
	file >> AxisZ; 
	if (!getSet(AxisZ, &(prefFileSet.dAngleZ))) bErr = true;
	file >> FactorX; 
	if (!getSet(FactorX, &(prefFileSet.dFactorX))) bErr = true;
	file >> FactorY; 
	if (!getSet(FactorY, &(prefFileSet.dFactorY))) bErr = true;
	file >> FactorZ; 
	if (!getSet(FactorZ, &(prefFileSet.dFactorZ))) bErr = true;
	file >> SegmentsQuantity; 
	if (!getSet(SegmentsQuantity, &(prefFileSet.iSegQty))) bErr = true;
	file >> SelectElements; 
	if (!getSet(SelectElements, &(prefFileSet.iSelElems))) bErr = true;
	file >> MirrorElements; 
	if (!getSet(MirrorElements, &(prefFileSet.iMirror))) bErr = true;
	file >> CreateNewDocument; 
	if (!getSet(CreateNewDocument, &(prefFileSet.iCreateNewDoc))) bErr = true;

	if (bErr)
		acutPrintf(_T("\n   - %s\n"), sIniFileName);

	file.close( );


}


/////////////////////////////////////////////////////////////////////
//----- ObjectARX EntryPoint
class CFormApp : public AcRxArxApp 
{

public:
	CFormApp () : AcRxArxApp () 
	{}

	///////////////////////////////////////////
	virtual AcRx::AppRetCode On_kInitAppMsg (void *pkt) 
	{
		AcRx::AppRetCode retCode =AcRxArxApp::On_kInitAppMsg (pkt) ;

		bSDI_changed = false;

		getSettings();

		if (checkSDI(false)) prefFileSet.iCreateNewDoc = 0;

		//iterateDb(acdbHostApplicationServices()->workingDatabase());

		return (retCode) ;
	}

	///////////////////////////////////////////
	virtual AcRx::AppRetCode On_kUnloadAppMsg (void *pkt) 
	{
		AcRx::AppRetCode retCode =AcRxArxApp::On_kUnloadAppMsg (pkt) ;
		
		//  
		if (dlg != NULL)
		{
			dlg->DestroyWindow();
			dlg = NULL;
		}
		if (sdlg != NULL)
		{
			sdlg->DestroyWindow();
			sdlg = NULL;
		}

		//  SDI  
		if (bSDI_changed)
		{
			struct resbuf rb;
			acedGetVar(_T("SDI"), &rb);
			rb.resval.rint = 1;
			acedSetVar(_T("SDI"), &rb);
		}

		return (retCode);
	}

	///////////////////////////////////////////
	virtual void RegisterServerComponents () 
	{
	}

	// ----- AKS._AKSSET command
	static void AKS_AKSSET(void)
	{

		acedUpdateDisplay();

		HWND hwndAcad = adsw_acadMainWnd();
		CWnd *pWnd = CWnd::FromHandle ( hwndAcad );

		if( sdlg == NULL ) 
		{
			CAcModuleResourceOverride resOverride;

			if ( (sdlg = new SetDlg ( pWnd )) == NULL )  return;

			BOOL succeeded = sdlg->Create( pWnd );
			if ( !succeeded )
				acutPrintf(_T("\n   \n"));
		}

		//  
		if (sdlg != NULL) 
		{
			sdlg->ShowWindow(SW_SHOWNORMAL);
			sdlg->syncToSettings();
		}
		else
			acutPrintf(_T("\n   !\n"));

	}

	///////////////////////////////////////////
	// ----- AKS._AKSPRO command
	static void AKS_AKSPRO(void)
	{
		if (!checkSDI()) return;

		acedUpdateDisplay();

		HWND hwndAcad = adsw_acadMainWnd();
		CWnd *pWnd = CWnd::FromHandle ( hwndAcad );

		if( dlg == NULL ) 
		{
			CAcModuleResourceOverride resOverride;

			if ( (dlg = new Dlg ( pWnd )) == NULL )  return;

			BOOL succeeded = dlg->Create( pWnd );
			if ( !succeeded )
				acutPrintf(_T("\n   \n"));
		}

		//  
		if (dlg != NULL) 
		{
			dlg->ShowWindow(SW_SHOWNORMAL);
			dlg->syncToSettings();
		}
		else
			acutPrintf(_T("\n   !\n"));
	}

	////////////////////////////////////////
	// ----- ads_akspro command
	static int ads_akspro(void)
	{
		struct resbuf *pArgs =acedGetArgs () ;

		if (!checkSDI(false)) 
		{
			acutPrintf(_T("\n AKSPRO    SDI = 0\n"));
			return RTERROR;
		}

		syncCurSetToFileSet();

		if (
			goAks(
			pArgs,  // 
			true, //   
			0  //   
			)
			) return RTERROR;

		reportNotProcessed();

		acedRetVoid() ;

		return (RSRSLT) ;
	}

	/////////////////////////////////////////////
	// ----- ads_akssel symbol (do not rename)
	static int ads_akssel(void)
	{
		//----- Remove the following line if you do not expect any argument for this ADS function
		struct resbuf *pArgs =acedGetArgs ();

		if (!checkSDI(false)) 
		{
			acutPrintf(_T("\n AKSSEL    SDI = 0\n"));
			return RTERROR;
		}

		syncCurSetToFileSet();
		
		if (
			goAks(
			pArgs,  // 
			true,  //   
			1 //  
			)
			) return RTERROR;

		reportNotProcessed();

		acedRetVoid ();

		return (RSRSLT);
	}

	/////////////////////////////////////////////
	// ----- ads_akssdi symbol (do not rename)
	static int ads_akssdi(void)
	{
		//----- Remove the following line if you do not expect any argument for this ADS function
		struct resbuf *pArgs =acedGetArgs ();

		syncCurSetToFileSet();

		if (
			goAks(
			pArgs, // 
			false //    
			)
			) return RTERROR;

		reportNotProcessed();

		acedRetVoid ();

		return (RSRSLT);
	}

	/////////////////////////////////////////////
	// ----- ads_akspic symbol (do not rename)
	static int ads_akspic(void)
	{
		//----- Remove the following line if you do not expect any argument for this ADS function
		struct resbuf *pArgs =acedGetArgs ();

		syncCurSetToFileSet();

		if (
			goAks(
			pArgs // 
			)
			) return RTERROR;

		reportNotProcessed();

		acedRetVoid ();

		return (RSRSLT);
	}


	// ----- AKS._AKS command (do not rename)
	static void AKS_AKS(void)
	{
		syncCurSetToFileSet();
		goAks();
		reportNotProcessed();
	}
} ;

////////////////////////////////
void printVers()
{
	acutPrintf(_T("\n//////////////////////////////////////////////////////////////////////////\n")
	           _T("      2.317-69 ( 1990.)\n")
	           _T("   ,    ,\n")
	           _T("        ;\n")
	           _T("    ,\n")
	           _T("        .\n")
	           _T(" - %s\n   2005   -  , .-\n")
             _T(" 2013   , \n")
	           _T(" - http://microstation.nm.ru/acad/aks/\n")
	           _T("  - wleo@yandex.ru\n"), vers);
	acedPostCommandPrompt();
}


//-----------------------------------------------------------------------------
IMPLEMENT_ARX_ENTRYPOINT(CFormApp)

ACED_ARXCOMMAND_ENTRY_AUTO(CFormApp, AKS, _AKSPRO, AKSPRO, ACRX_CMD_TRANSPARENT, NULL)
ACED_ARXCOMMAND_ENTRY_AUTO(CFormApp, AKS, _AKSSET, AKSSET, ACRX_CMD_TRANSPARENT, NULL)
ACED_ARXCOMMAND_ENTRY_AUTO(CFormApp, AKS, _AKS, AKS, ACRX_CMD_TRANSPARENT, NULL)

//ACED_ARXCOMMAND_ENTRY_AUTO(CFormApp, AKS, _AKSPRO, AKSPRO, ACRX_CMD_SESSION | ACRX_CMD_TRANSPARENT, NULL)
ACED_ADSCOMMAND_ENTRY_AUTO(CFormApp, akspro, true)
ACED_ADSCOMMAND_ENTRY_AUTO(CFormApp, akssel, true)
ACED_ADSCOMMAND_ENTRY_AUTO(CFormApp, akssdi, true)

ACED_ADSSYMBOL_ENTRY_AUTO(CFormApp, akspic, true)

/*
^C^C(arxload "W_AksPro.arx") _akspro
(c:akspro 1)
(c:akspro 0. 45. 90.)
*/

